;;;
;;; MemoryAllocate.s
;;;
;;; created: <19.10.1998>
;;; last modified: <Wed Apr 26 17:00:53 2000>
;;;
;;; Christian Fowelin <christian@fowelin.de>
;;;
;;; Released under the LGPL (file COPYING).
;;;
;;; This is to handle memory allocation in assembly much easier.
;;;

;;; Sets up a memory managment system which uses chunks. If the user
;;; application destroys them, this system will fail and do sys_exit(1).

;;; If the application wants to free allocated memory which is mapped, it
;;; should unmap it first!

;;; The code is self-explaining :-)

%include "system.h"
%include "mem.h"
%include "io.h"

global _MemoryAllocate

extern _errno,_memory_brk,_memory_firstchunk,_memory_LOCK,_memory_HAS_INIT
extern _MemoryInit,_MemoryAllowReadWriteHeaders,_MemoryRestoreHeaders
extern _MemoryTune,_MemoryChunkStructureIsCorrupted

	section .text
	align 32

;;;
;;; MemoryAllocate
;;;
;;;  In: MemoryAllocate size (eax)
;;;  Out: eax=ptr (on success); -1 (on error)
;;;
;;; a=mem to alloc+Header
;;; b(0..1)=0 (tmp: 0:length, 1:addr)
;;; Scan for free chunk
;;;  last Chunk: b(0..1)=0? --> get more brk
;;;  Found: >=a  ; < b(0..1) ; -> b(0..1)
;;; Make Chunk (+opt: free rest...)
_MemoryAllocate:
	__push ebx, ecx, edx, esi, edi

	push eax
	MemoryHAS_INIT		; '_MemoryAllocate: '
	MemoryLOCK
	pop eax

	add eax,_memory_chunk_size	; Usermemory is chunk.size-chunk_size
	add eax,PAGE_SIZE-1
	and eax,PAGE_MASK
	mov ebx,0xffffffff
	mov ecx,0

	call _MemoryAllowReadWriteHeaders
	mov esi,[_memory_firstchunk]
.SearchLoop:
	;; Check if ID is OK, if not, exit an give ERROR!
	mov edx,[esi+_memory_chunk.flags]
	and edx,_memory_flag_IDmask
	cmp edx,_memory_identify_flag
	je .SearchLoop_j1
	mov eax,dword .rout_name
	jmp _MemoryChunkStructureIsCorrupted
.SearchLoop_j1:
	;; Last chunk?
	mov edx,[esi+_memory_chunk.flags]
	and edx,_memory_lastchunk_flag
	jnz .WasLastChunk
	;; No, is it the first chunk?
	mov edx,[esi+_memory_chunk.flags]
	and edx,_memory_firstchunk_flag
	jnz .ScanForNextChunk
	;; No, is it a free chunk?
	mov edx,[esi+_memory_chunk.flags]
	and edx,_memory_freechunk_flag
	jz .ScanForNextChunk
	;; Valid? (Size)
	cmp eax,[esi+_memory_chunk.size]
	ja .ScanForNextChunk
	;; Yes, is free so test, if smaller than prev., save it
	cmp ebx,[esi+_memory_chunk.size]
	jbe .ScanForNextChunk	; jbe: allocated chunk will be located in the first valid chunk
	;; Yes: 'mark'
	mov ebx,[esi+_memory_chunk.size]
	mov ecx,esi

.ScanForNextChunk:
	;; Find next
	add esi,[esi+_memory_chunk.size]
	jmp .SearchLoop

.WasLastChunk:
	;; esi=ptr to last chunk
	;; The last chunk is allways allocated, so we need to get more memory, if ecx=0
	jecxz .AllocNewBrk
	jmp .AllocFreeChunk

.AllocNewBrk:
	;; brk=brk+(eax+(PAGE_SIZE-1))and PAGE_MASK)
	mov edx,eax	; !!
	add eax,[_memory_brk]
	sys_brk eax
	;; Error?
	cmp eax,0
	jnle .AllocNewBrkOK

	;; Return Error
	mov [_errno],eax
	call _MemoryRestoreHeaders
	MemoryUNLOCK
	__pop edi, esi, edx, ecx, ebx
	mov eax,-1
	ret

; No brk-error
.AllocNewBrkOK:
; Create Header (Flags, size of prev, size)
	mov edi,[_memory_brk]
	mov [_memory_brk],eax
	mov [edi+_memory_chunk.size],edx
	mov eax,[esi+_memory_chunk.size]
	mov [edi+_memory_chunk.prev],eax
	mov [edi+_memory_chunk.flags],dword _memory_lastchunk_flag|_memory_identify_flag|_memory_PROT_READ|_memory_PROT_WRITE
	mov [edi+_memory_chunk.proc],dword 0
; Get pid
	sys_getpid
	mov [edi+_memory_chunk.pid],eax

; Mark prv as non-last
	and [esi+_memory_chunk.flags],dword ~(_memory_lastchunk_flag)

; MemoryProtect
	sys_mprotect edi, dword [edi+_memory_chunk.size], dword PROT_READ|PROT_WRITE
; Return Pointer in edi
	jmp .GoodEnd

.AllocFreeChunk:
;1)
; Size of chunk=(alloc+(PAGE_SIZE-1))and PAGE_MASK)
; Write Header
;  size of free - size of chunk >= PAGE_SIZE ; then a=true
;  if a
;   size=size of chunk
;  else
;   size=size of free
;  endif
;   get prev. size (should allready be initialized)
;  mprotect (except 1st PAGE)
;2)
; if a
;  Write free Header
;   size=size of free chunk - size of chunk
;   mprotect
; endif
;3)
; if a     (should be initialized for free chunk, so no need to mod. if
;           new chunk fits space)
;  Modify next header
;  prev. size=size of free chunk - size of chunk
; endif

	mov edi,ecx
; We need to read/write the hole free chunk
	push eax
	sys_mprotect edi, dword [edi+_memory_chunk.size], dword PROT_READ|PROT_WRITE
	pop eax
	mov edx,[edi+_memory_chunk.size]
	sub edx,eax
	setnz cl

	mov [edi+_memory_chunk.flags],dword _memory_identify_flag|_memory_PROT_READ|_memory_PROT_WRITE
; prev is OK!
; Return Pointer
	push edi

	cmp cl,1
	jne near .AllocFreeChunk_j1
	mov [edi+_memory_chunk.size],eax
	add edi,eax
	mov [edi+_memory_chunk.prev],eax
	mov [edi+_memory_chunk.size],edx
	mov [edi+_memory_chunk.flags],dword _memory_freechunk_flag|_memory_identify_flag|_memory_PROT_NONE
	mov [edi+_memory_chunk.proc],dword 0
	sys_mprotect edi, edx, dword PROT_NONE
	sys_mprotect edi, dword PAGE_SIZE, dword PROT_READ|PROT_WRITE
	add edi,edx
	mov [edi+_memory_chunk.prev],edx

.AllocFreeChunk_j1:
	pop edi
; Get pid
	sys_getpid
	mov [edi+_memory_chunk.pid],eax

.GoodEnd:
	call _MemoryRestoreHeaders
	MemoryUNLOCK
	mov eax,edi
	add eax,_memory_chunk_size
	__pop edi, esi, edx, ecx, ebx
; Tune Memory!
	call _MemoryTune
	ret

	section .data
.rout_name:	db '_MemoryAllocate',0



;;; Local Variables:
;;;  mode:asm
;;;  eval:(load-library "time-stamp")
;;;  eval:(make-local-variable 'write-file-hooks)
;;;  eval:(add-hook 'write-file-hooks 'time-stamp)
;;;  eval:(setq time-stamp-start "last modified:[ 	]+\\\\?[\"<]+")
;;;  eval:(setq time-stamp-format "%3a %3b %2d %02H:%02M:%02S %y")
;;; End: