﻿/*
 * Created by SharpDevelop.
 * User: de8245
 * Date: 12/29/2007
 * Time: 8:46 PM
 * 
 * To change this template use Tools | Options | Coding | Edit Standard Headers.
 */
using System;
using System.IO;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Drawing;
using System.Reflection;
using System.Diagnostics;
using System.Runtime.CompilerServices;
using System.Threading;
using System.Globalization;
using VizBin.util;
using VizBin.searcher;

namespace VizBin
{
	class Controller
	{
        static long XL_MAX_HEIGHT = 65535;
        // the input file name we're going to read from
        private String is_input_file;
		// the 'row" in the table we're currently referencing
		private int in_row = 0;
		// the 'column' in the table we're currently referencing
		private int in_col = 0;
		// the table we're using fo the image
		private OffsetTable IO_TABLE = new OffsetTable();
        // mapping of names of Image processors to instances of them
        private static IDictionary<String, AbstractImageProcessor> xo_names2procs = new Dictionary<String, AbstractImageProcessor>();
		// the size of the image that is to be rendered
		private Size io_size = new Size( 800, 600 );
        // the file parser to use -- magic file --> for searching
        private MagicFileParser io_parser = MagicFileParser.getParser();
        // the scale currently being used
        private int in_scale = 1;

		/// <summary>
		/// Default constructor preventing public instantiation
		/// </summary>
		internal Controller()
		{
			// there is nothing to do here at the moment
		}

		/// <summary>
		/// Updates the size of the image we are going to render. This method will also force
		/// the OffsetTable to be re-initialized if there is a file already specified.
		/// </summary>
		/// <param name="vo_image_size">the size of the image to generate</param>
		internal void setImageSize( Size vo_image_size )
		{
			if (vo_image_size == null)
				throw new NullReferenceException("Size object cannot be null");
			this.io_size = new Size(vo_image_size.Width, vo_image_size.Height);
			if (this.is_input_file != null)
				this.initTable();
		}

        /// <summary>
        /// Returns true if the filename has been set, false otherwise
        /// </summary>
        /// <returns>true if set, false otherwise</returns>
        internal bool isInputFileSet()
        {
            return this.is_input_file != null;
        }

        /// <summary>
        /// Sets the input filename that will be used for image processing. If the file is not valid
        /// an error message is displayed for the user. True is returned if the file is set, false otherwise.
        /// </summary>
        /// <returns>true if set, false otherwise</returns>
        /// <param name="vs_filename"></param>
        internal bool setInputFilename(String vs_filename)
        {
            bool is_set = false;
            if (Controller.isFileExists(vs_filename))
            {
                this.is_input_file = vs_filename;
                this.in_col = this.in_row = 0;
				is_set = this.initTable();
            }
            else
                Controller.displayErrorMessage("Invalid File Name Selected. Please select a valid file.");
            return is_set;
        }

		/// <summary>
		/// Initialize the internal byte offset table with the appropriate values.
		/// </summary>
		private bool initTable()
		{
			if (this.is_input_file == null)
				throw new ArgumentException("Filename is null");
			if (this.io_size == null)
				throw new ArgumentException("Size is null");
            bool b_done = false;

			// calculate roughly how many bytes per bite to read
			long n_bytes_to_read = this.io_size.Width * this.io_size.Height;
			//long l_rem_bits = n_bytes_to_read % 8;
			//n_bytes_to_read = (l_rem_bits == 0 ? (n_bytes_to_read / 8) : (n_bytes_to_read / 8) + 1);
			
			// now make sure the file size isn't less than the number of bytes we want to read
			// if it is, adjust the number of bytes to read
			FileInfo o_finfo = new FileInfo(this.is_input_file);
			if (o_finfo.Length < n_bytes_to_read)
				n_bytes_to_read = (int)o_finfo.Length;

            try
            {
                IO_TABLE.initTable(n_bytes_to_read, o_finfo.Length);
                b_done = true;
            }
            catch (ArgumentException vo_e)
            {
                Controller.displayErrorMessage("Cannot display image.\n" + vo_e.Message);
            }
            return b_done;
		}

        /// <summary>
        /// Get an image processor instance 
        /// </summary>
        /// <param name="vs_proc_name"></param>
        /// <returns></returns>
        internal AbstractImageProcessor getImageProcessor(String vs_proc_name)
        {
            if (!Controller.isValidString(vs_proc_name))
                throw new ArgumentException("Inavlid name, name is null or empty");
            if(!(xo_names2procs.ContainsKey(vs_proc_name)))
                throw new ArgumentException("Processor for name doesn't exist");
            return xo_names2procs[vs_proc_name];
        }

        /// <summary>
        /// Return an array of all of the image processor names currently in the assembly. The first time
        /// the method is called it will init the name to instance mapping of all processors.
        /// </summary>
        /// <returns>array of string names</returns>
        internal static String[] getImageProcessorNames()
        {
            String[] as_names = null;
            if ( xo_names2procs.Keys.Count < 1 )
            {
                Process o_proc = Process.GetCurrentProcess();
                Type o_type = typeof(AbstractImageProcessor);
                Assembly o_asm = Assembly.GetAssembly(o_type);
                ICollection<Type> o_sub_classes = Controller.ListSubClassesForType(o_asm, o_type);

                // init each subclass
                if (o_sub_classes != null)
                {
                    Type[] o_empty_types = new Type[0];
                    foreach (Type o_sub_type in o_sub_classes)
                    {
                        ConstructorInfo o_const_info = o_sub_type.GetConstructor(o_empty_types);
                        AbstractImageProcessor o_img_proc = (AbstractImageProcessor)o_const_info.Invoke(null);
                        xo_names2procs.Add(o_img_proc.getName(), o_img_proc);
                    }
                }
            }
            as_names = xo_names2procs.Keys.ToArray<String>();
            return as_names;
        }

        /// <summary>
        /// Returns an enumeration of all classes in the given assembly that are subclasses of the
        /// given base type.
        /// </summary>
        /// <param name="asm">The assembly to use</param>
        /// <param name="baseclassType">The base class to seach with</param>
        /// <returns>Enumeration of Types</returns>
        private static ICollection<Type> ListSubClassesForType(Assembly asm, Type baseclassType)
        {
            if (asm == null)
				throw new ArgumentException("Assembly is null");
            if (baseclassType == null)
				throw new ArgumentException("base class type is null");
            ICollection<Type> o_types = new List<Type>();
            Type[] ao_types = asm.GetTypes();
            foreach (Type o_type in ao_types)
            {
                if (o_type.IsSubclassOf(baseclassType))
                    o_types.Add(o_type);
            }
            return o_types;
        }

		/// <summary>
		/// Returns a FileStream which allows writing as well as reading. If the file path parameter is invalid
		/// (null or the empty string) an ArgumentException is thrown.
		/// </summary>
		/// <param name="vs_filepath"></param>
		/// <returns></returns>
		internal static FileStream getRWFileStream(String vs_filepath)
		{
			FileStream o_fs = null;
			if (Controller.isValidString(vs_filepath))
			{
				bool b_write = false;
				if (File.Exists(vs_filepath))
					if (Controller.displayMessage("File Exists. Do you wish to overwrite?", "Confirm File Overwrite", MessageBoxButtons.OKCancel) == DialogResult.OK)
						b_write = true;
					else
						b_write = true;

				if (b_write)
				{
					try
					{
						o_fs = File.Open(vs_filepath, FileMode.Create, FileAccess.ReadWrite);
					}
					catch (Exception vo_e)
					{
						Controller.displayErrorMessage("Unable to open file for writing.\n" + vo_e.Message);
					}
				}
			}
			else throw new ArgumentException("File path is invalid, null, or empty");
			return o_fs;
		}

		/// <summary>
		/// Given the String, return a READ ONLY file stream. If the path is null or 
		/// empty, an ArgumentException will be thrown. If the file doesn't exist, an Argument
		/// Exception is thrown. If there is an error reading from the file, an error message
		/// will be displayed and a null file stream will be retruned.
		/// </summary>
		/// <param name="vs_filepath">String full path to the file to be read</param>
		/// <returns>The FileStream to read from</returns>
		internal static FileStream getROFileStream(String vs_filepath)
		{
			FileStream o_fs_out = null;
			if( Controller.isValidString( vs_filepath ))			
			{
				if( File.Exists( vs_filepath ))
				{
					try
					{
						o_fs_out = File.OpenRead(vs_filepath);
					}
					catch (IOException vo_ioex)
					{
						Controller.displayErrorMessage("Unable to read file.\n" + vo_ioex.Message);	
					}
				}
				else throw new ArgumentException("File: " + vs_filepath + " does not exist!");
			}
			else throw new ArgumentException("FilePath given is either null or empty");
			return o_fs_out;
		}

		/// <summary>
		/// Given a filename and the offset to read out of it, return the byte array
		/// represented by the offset locations.
		/// </summary>
		/// <param name="vs_filename">the file to read</param>
		/// <param name="vo_struct">the offset to read out of</param>
		/// <returns>byte array</returns>
		/// <exception cref="ArgumentException">if the file string is not valid, or the offset struct is null</exception>
		private static byte[] getArrayAtOffset(String vs_filename, OffsetStruct vo_struct)
		{
			if (vo_struct == null) throw new ArgumentException("OffsetStruct cannot be null");
			if (!Controller.isValidString(vs_filename)) throw new ArgumentException("Invalid File: " + vs_filename);
			byte[] ab_bytes = new byte[vo_struct.getEnd() - vo_struct.getStart()];
			FileStream o_fs = Controller.getROFileStream(vs_filename);
			if (o_fs != null && o_fs.CanRead)
			{
				o_fs.Position = vo_struct.getStart();
				o_fs.Read(ab_bytes, 0, ab_bytes.Length);
			}
			Controller.close(o_fs);
			return ab_bytes;
		}

		/// <summary>
		/// Create's an overview map by reading some of the bytes from the file without sending
		/// the entire file through the image processor. If the file length is smaller than the 
		/// number of bytes that can be represented in the 
		/// </summary>
		/// <param name="vn_width"></param>
		/// <param name="vn_height"></param>
		/// <param name="vo_img_proc"></param>
		/// <returns></returns>
		internal Image getBinaryMap(int vn_width, int vn_height, AbstractImageProcessor vo_img_proc)
		{
			if (vn_width < 1 || vn_width > XL_MAX_HEIGHT)
				throw new ArgumentException("Width must be between 1 and 655355. Invalid width: " + vn_width);
			else if (vn_height < 1 || vn_height > XL_MAX_HEIGHT)
				throw new ArgumentException("Height must be between 1 and 65535. Invalid height: " + vn_height);
			else if (vo_img_proc == null)
				throw new ArgumentException("Image processor cannot be null");

            long n_bytes_to_read = vn_height * vn_width;
			FileInfo o_info = new FileInfo(this.is_input_file);
			
			// shorten the reading array of bytes to match the file length if the
			// actual height and width gives us too many bytes to read
			if (o_info.Length < n_bytes_to_read)
				n_bytes_to_read = o_info.Length;
			
			// determine the skip value (every N bytes to read) when the byte array length is
			// smaller than the file, otherwise read every 1 byte - avoid the dreaded divide by 0
			// error here as well
			long n_skip_value = 1;
			if (n_bytes_to_read < o_info.Length && o_info.Length > 0 )
				n_skip_value = o_info.Length / n_bytes_to_read;

			// fill our byte array with bytes and send it to the image processor
			byte[] ab_bytes = new byte[n_bytes_to_read];
			Image o_img = null;
			FileStream o_fs = Controller.getROFileStream(this.is_input_file);
			if (o_fs != null )
			{
				long l_pos = o_fs.Position;
				int n_byte_idx = 0;
				while ((l_pos <= o_fs.Length) && (n_byte_idx < n_bytes_to_read) && o_fs.CanRead)
				{
					ab_bytes[n_byte_idx] = (byte)o_fs.ReadByte();
					n_byte_idx++;
					l_pos = l_pos + n_skip_value;
					o_fs.Position = l_pos;
				}
				Controller.close(o_fs);
			}
			o_img = vo_img_proc.getOverviewImage(ab_bytes, vn_width, vn_height + 1);
			return o_img;
		}

        /// <summary>
        /// Returns true if we can move down in the memory table
        /// </summary>
        /// <returns>bool</returns>
        internal bool canMoveDown()
        {
            bool b_retval = false;
            if(IO_TABLE != null && IO_TABLE.getRowCount() -1 > this.in_row)
                b_retval = IO_TABLE.getCell(this.in_row + 1, this.in_col ) != null;
            return b_retval;
        }

        /// <summary>
        /// Returns true if we can move up in the memory table
        /// </summary>
        /// <returns>bool</returns>
        internal bool canMoveUp()
        {
            bool b_retval = false;
            if(IO_TABLE != null && this.in_row > 0)
                b_retval = IO_TABLE.getCell(this.in_row - 1, this.in_col ) != null;
            return b_retval;
        }

        /// <summary>
        /// Returns true if we can move left in the memory table
        /// </summary>
        /// <returns>bool</returns>
        internal bool canMoveLeft()
        {
            bool b_retval = false;
            if(IO_TABLE != null && this.in_col > 0)
                b_retval = IO_TABLE.getCell(this.in_row, this.in_col -1) != null;
            return b_retval;
        }

        /// <summary>
        /// Returns true if we can move right in the memory table
        /// </summary>
        /// <returns>bool</returns>
        internal bool canMoveRight()
        {
            bool b_retval = false;
            if (IO_TABLE != null && IO_TABLE.getColumnCount(this.in_row) -1 > this.in_col)
                b_retval = IO_TABLE.getCell(this.in_row, this.in_col + 1) != null;
            return b_retval;
        }

        /// <summary>
        /// Returns an image to be displayed in the display box. Throws ArgumentExceptions if
        /// the image processor is null, the height/width is less than 1 or greater than 65535,
        /// </summary>
        /// <returns>Image</returns>
		/// <param name="ve_direction">The direction in the table of indexes to move</param>
        /// <param name="vo_img_proc">the subclass of AbstractImageProcessor used to render the image</param>
        /// <exception cref="ArgumentException">vo_img_proc is null</exception>
        /// <exception cref="ArgumentException">height/width less than 1 or greater than 65535</exception>
        internal Image getImage(AbstractImageProcessor vo_img_proc, DirectionalEnum ve_direction)
        {
            if (vo_img_proc == null)
                throw new ArgumentException("Image Processor cannot be null");

			Image o_img = null;
			switch (ve_direction)
			{
				case DirectionalEnum.CENTER:
					break;
				case DirectionalEnum.DOWN:
					if (IO_TABLE.getRowCount() -1> this.in_row)
						this.in_row++;
					break;
				case DirectionalEnum.LEFT:
					if( this.in_col > 0 )
						this.in_col--;
					break;
				case DirectionalEnum.RIGHT:
					if( IO_TABLE.getColumnCount(this.in_row) - 1 > this.in_col )
						this.in_col++;
					break;
				case DirectionalEnum.UP:
					if( this.in_row > 0 )
						this.in_row--;
					break;
				default:
					throw new ArgumentException("Unrecognized Directional Enum: " + ve_direction);
			}

			OffsetStruct o_struct = IO_TABLE.getCell(this.in_row, this.in_col);
            o_img = Controller.getImage(this.is_input_file, vo_img_proc, this.io_size, o_struct);
            // update the scaling factor
            this.in_scale = AbstractImageProcessor.getScalingFactor(this.io_size.Width, this.io_size.Height, Convert.ToInt32(o_struct.getEnd() - o_struct.getStart()));
			return o_img;
        }

        /// <summary>
        /// Returns all of the search results for the given string
        /// </summary>
        /// <param name="vs_search">the string to search for</param>
        /// <returns>the string containing all of the search results</returns>
        internal String getSearchResults(String vs_search)
        {
            //determine if the sting is a hex string
            //if it is, then parse it and search for it
            //otherwise abort and search for the ascii string

			// get the byte array we're currently looking at
			byte[] ab_curr_image = Controller.getArrayAtOffset( this.is_input_file, this.IO_TABLE.getCell( this.in_row, this.in_col));

			// determine if the requested search is for a string or actual byte(s)
			bool b_is_byte = false;
			byte[] ab_bytes = null;

			try
			{
				byte b_tmp_byte = Convert.ToByte(vs_search, 16);
				ab_bytes =  new byte[]{ b_tmp_byte };
				b_is_byte = true;
			}
			catch (Exception)
			{
				byte b_tmp_byte = 0;
				if (Byte.TryParse(vs_search, out b_tmp_byte))
				{
					ab_bytes = new byte[] { b_tmp_byte };
					b_is_byte = true;
				}
			}

			// convert the given string to a byte array to search for
			if (!b_is_byte)
			{
				char[] ac_chars = vs_search.ToCharArray();
				ab_bytes = new byte[ac_chars.Length];
				for (int i = 0; i < ac_chars.Length; i++)
					ab_bytes[i] = Convert.ToByte(ac_chars[i]);
			}

			// do the search
			DefaultSearch o_srch = new DefaultSearch();
			ICollection<SearchResult> o_res = o_srch.getResults(vs_search, ab_bytes, ab_curr_image);
			
			// build the result string 
			StringBuilder o_bldr = new StringBuilder("Search Results:");
			foreach (SearchResult sr in o_res)
			{
				o_bldr.Append(Environment.NewLine);
				o_bldr.Append(sr.ToString());
			}
			return o_bldr.ToString();
        }

        /// <summary>
        /// Gets the results from the magic file parser.
        /// </summary>
        /// <returns>String to be presented to the user</returns>
        internal String getMagicSearchResults()
        {
            OffsetStruct o_struct = this.IO_TABLE.getCell( this.in_row, this.in_col );
            
            //fill this array for the image processor
            byte[] ab_bytes = new byte[o_struct.getEnd() - o_struct.getStart()];

            //read from last offset to the new offset
            FileStream o_fs = Controller.getROFileStream(this.is_input_file);
			if (o_fs != null && o_fs.CanRead)
			{
				o_fs.Position = o_struct.getStart();
				o_fs.Read(ab_bytes, 0, ab_bytes.Length);
			}
            Controller.close(o_fs);
            ICollection<SearchResult> o_res = this.io_parser.parse(ab_bytes);
            StringBuilder o_bldr = new StringBuilder("Magic Search Results:\n");
			if (o_res.Count > 0)
				foreach (SearchResult sr in o_res)
					o_bldr.Append("\n" + sr.ToString());
			else
				o_bldr.Append("No Results for Current byte chunk");
            return o_bldr.ToString();
        }

        /// <summary>
        /// Returns a string which describes the current offset values
        /// </summary>
        /// <returns>string</returns>
        internal String getOffsetDisplayString()
        {
            StringBuilder o_sb = new StringBuilder("Bytes: ");
            OffsetStruct o_struct = this.IO_TABLE.getCell(this.in_row, this.in_col);
            if (o_struct != null)
            {
                o_sb.Append(o_struct.getStart());
                o_sb.Append(" to ");
                o_sb.Append(o_struct.getEnd());
                o_sb.Append(" of ");
                o_sb.Append(new FileInfo(this.is_input_file).Length);
            }
            return o_sb.ToString();
        }

        /// <summary>
        /// Returns a string representing the current image scaling factor
        /// used by the image rendering engines (if they chose to use it) or
        /// a null string if the scaling/zoom factor is 1
        /// </summary>
        /// <returns>the descriptive string representing the scaling factor</returns>
        internal String getScale()
        {
            return "Current Image Scaling Factor: " + this.in_scale + " pixels per byte";
        }

		/// <summary>
		/// Retruns a hex string based on the OffsetStruct in the cell in the OffsetStructTable
		/// defined by the direction the user would like to move in the table
		/// </summary>
		/// <param name="ve_direction">the direction in the table</param>
		/// <returns>string</returns>
		internal String getHex(DirectionalEnum ve_direction)
		{
			int n_row = this.in_row;
			int n_col = this.in_col;
			switch (ve_direction)
			{
				case DirectionalEnum.CENTER:
					break;
				case DirectionalEnum.DOWN:
					if (IO_TABLE.getRowCount() - 1 > n_row)
						n_row++;
					break;
				case DirectionalEnum.LEFT:
					if (n_col > 0)
						n_col--;
					break;
				case DirectionalEnum.RIGHT:
					if (IO_TABLE.getColumnCount(n_row) - 1 > n_col)
						n_col++;
					break;
				case DirectionalEnum.UP:
					if (this.in_row > 0)
						this.in_row--;
					break;
				default:
					throw new ArgumentException("Unrecognized Directional Enum: " + ve_direction);
			}
			OffsetStruct o_struct = IO_TABLE.getCell(n_row, n_col);
			String s_string = null;
			if (o_struct != null)
				s_string = Controller.getHex(this.is_input_file, this.io_size, o_struct);
			return s_string;
		}

		/// <summary>
		/// Create a byte string given the offset struct to read from.
		/// </summary>
		/// <param name="vs_filename">the file to read</param>
		/// <param name="vo_size">the size of the image that was generated</param>
		/// <param name="vo_struct">the offset struct used to generate the hex</param>
		/// <returns></returns>
		[MethodImpl(MethodImplOptions.Synchronized)]
		private static String getHex(String vs_filename, Size vo_size, OffsetStruct vo_struct)
		{
			if (vo_size == null)
				throw new ArgumentException("Size object cannot be null");
			if (vo_size.Height < 1 || vo_size.Height > XL_MAX_HEIGHT)
				throw new ArgumentException("Height must be between 1 and " + XL_MAX_HEIGHT);
			if (vo_size.Width < 1 || vo_size.Width > XL_MAX_HEIGHT)
				throw new ArgumentException("Width must be between 1 and " + XL_MAX_HEIGHT);
			if (vo_struct == null)
				throw new ArgumentException("OffsetStruct cannot be null");
			if (!Controller.isFileExists(vs_filename))
				throw new ArgumentException("String filename cannot be null or an invalid file");

			// create a byte array slice to read from
			byte[] ab_bytes = Controller.getArrayAtOffset(vs_filename, vo_struct);

			// turn the bytes into a 'string' and return them
			StringBuilder o_bldr = new StringBuilder();
			StringBuilder o_hex_bldr = new StringBuilder();
			StringBuilder o_ascii_bldr = new StringBuilder();

			//iterate over the bytes
			long l_start_byte = vo_struct.getStart();
			int n_byte_counter = 0;
			
			//calculate the text length for column 1
			int n_length = vo_struct.getEnd().ToString().Length;
			long l_col1_length = n_length * 2;
			l_col1_length = l_col1_length + 4;
			int n_col1_length = Convert.ToInt32(l_col1_length);
			
			// this will build the appropriate zero formatter for the byte offset column
			StringBuilder s_zero_format_builder = new StringBuilder();
			for (int n = 0; n < n_length; n++)
				s_zero_format_builder.Append("0");

			for (int i = 0; i < ab_bytes.Length; i++)
			{
				// display hex
				byte b = ab_bytes[i];
				o_hex_bldr.Append(Convert.ToString(b, 16).PadLeft(2, '0'));
				o_hex_bldr.Append(' ');

				// display ascii
				char c_char = Convert.ToChar(b);
				if (!Char.IsLetterOrDigit(c_char) && !Char.IsPunctuation(c_char))
					c_char = '.';
				o_ascii_bldr.Append(c_char);
				o_ascii_bldr.Append(' ');

				n_byte_counter++;
				if (n_byte_counter >= 7 && i + 1 < ab_bytes.Length)
				{
					// col 1
					string s_bytecol = (l_start_byte + 1).ToString(s_zero_format_builder.ToString()) 
						+ " - " + (l_start_byte + n_byte_counter).ToString(s_zero_format_builder.ToString());
					o_bldr.Append(s_bytecol.PadRight(n_col1_length, ' '));
					l_start_byte = l_start_byte + n_byte_counter;

					// col 2
					o_bldr.Append(o_hex_bldr.ToString().PadLeft(25, ' '));
					o_hex_bldr = new StringBuilder();

					// col 3
					o_bldr.AppendLine(o_ascii_bldr.ToString().PadLeft(20, ' '));
					o_ascii_bldr = new StringBuilder();

					n_byte_counter = 0;
				}
			}

			return o_bldr.ToString();

			//int count = 0;
			//foreach (Byte b in ab_bytes)
			//{
			//    o_bldr.Append(Convert.ToString(b, 16).PadLeft(2, '0'));
			//    o_bldr.Append(' ');

			//    char c_char = Convert.ToChar(b);
			//    if (!Char.IsLetterOrDigit(c_char))
			//        c_char = '.';
			//    o_char_bldr.Append('.');
			//    o_char_bldr.Append(c_char);
			//    o_char_bldr.Append(' ');

			//    if (count == 7)
			//    {
			//        o_bldr.Append(Environment.NewLine);
			//        o_char_bldr.Append(Environment.NewLine);
			//        count = 0;
			//    }
			//    else
			//        count++;
			//}

			//String[] as_return_array = new String[2];
			//as_return_array[0] = o_bldr.ToString();
			//as_return_array[1] = o_char_bldr.ToString();
			//return as_return_array;
		}

        /// <summary>
        /// Uses the specified (TO BE ADDED) parser to generate one or more images to display based
        /// on the binary stream of values from the file specified.
        /// </summary>
        /// <param name="vs_filename">filename to read</param>
        /// <param name="vo_img_proc">Image Processor to use to create the image</param>
		/// <param name="vo_size">Size of the image to create</param>
		/// <param name="vo_struct">Offset struct which contains the start and end offsets to read</param>
		/// <returns>Image</returns>
        /// <exception cref="ArgumentException">if the height/width less than 1 or greater than 65535</exception>
        /// <exception cref="ArgumentException">if the filename is invalid or the file doesnt exist</exception>
		/// <exception cref="NullReferenceException">if the image processor subclass is null</exception>
		/// <exception cref="NullReferenceException">if the offset struct is null</exception>
		/// <exception cref="NullReferenceException">if the size object is null</exception>
		[MethodImpl(MethodImplOptions.Synchronized)]
        private static Image getImage(String vs_filename, AbstractImageProcessor vo_img_proc, Size vo_size, OffsetStruct vo_struct)
        {
			if (vo_size == null)
				throw new NullReferenceException("Size object cannot be null");
			if (vo_size.Width <= 0 || vo_size.Width > XL_MAX_HEIGHT)
				throw new ArgumentException("Width: " + vo_size.Width + " is invalid. Width must be between 1 and 65535");
			if (vo_size.Height <= 0 || vo_size.Height > XL_MAX_HEIGHT)
				throw new ArgumentException("Height: " + vo_size.Height + " is invalid. Height must be between 1 and 65535");
			if (!isFileExists(vs_filename))
				throw new ArgumentException("File name: " + vs_filename + " is invalid. The file path is invalid or the file does not exist.");
			if (vo_img_proc == null)
				throw new NullReferenceException("Image processor cannot be null");
			if (vo_struct == null)
				throw new NullReferenceException("Offset Struct cannot be null");

			Image o_img = null;

            //fill this array for the image processor
			byte[] ab_bytes = Controller.getArrayAtOffset(vs_filename, vo_struct);

            // send the byte array to the processor
			o_img = vo_img_proc.getImage(ab_bytes, vo_size.Width, vo_size.Height+1);

            return o_img;
        }

        /// <summary>
        /// Determines whether or not a file exists. First, the string used as the file path 
        /// is checked to make sure it is a correct string. Secondly,  the File.Exists method
        /// is called to determine whether or not the file exists. This is designed to be used
        /// instead of the File.Exists method by itself. Also, this will verify we can read from the
		/// file in addition to the file existance.
        /// </summary>
        /// <param name="vs_file_path">The path to the file</param>
        /// <returns>true if the file exists, false otherwise</returns>
        private static bool isFileExists(String vs_file_path)
        {
            bool b_valid = false;
            if (Controller.isValidString(vs_file_path))
            {
                if (File.Exists(vs_file_path))
                    try {
                        using( FileStream o_fs = new FileStream(vs_file_path, FileMode.Open ))
                        {
                            b_valid = o_fs.CanRead;
                        }
                    }
                    catch(Exception)
                    {
                        b_valid = false;
                    }
            }
            return b_valid;
        }

		/// <summary>
		/// Returns true if the string is non-null and not-empty (does not eval to "" when trim() is used).
		/// </summary>
		/// <param name="vs_string_to_test">The string to test</param>
		/// <returns>true if the string is valid, false otherwise</returns>
		private static bool isValidString(String vs_string_to_test)
		{
			bool b_valid = true;
			if (vs_string_to_test == null)
				b_valid = false;
			else
			{
				String s_test = vs_string_to_test.Trim();
				if (s_test.Length == 0)
					b_valid = false;
			}
			return b_valid;
		}

        /// <summary>
        /// Displays an error message with the given text and does
        /// a data dump on the exception
        /// </summary>
        /// <param name="vs_message">the message to display</param>
        /// <param name="vo_e">the exception to parse</param>
        /// <exception cref="ArgumentException">thrown if either parameter is null</exception>
        internal static void displayExcepton(String vs_message, Exception vo_e)
        {
            if (vo_e == null)
                throw new ArgumentException("Exception cannot be null (how ironic)");
            if (vs_message == null)
                throw new ArgumentException("Message to display cannot be null");
            StringBuilder o_builder = new StringBuilder(vs_message);
            o_builder.Append('\n');
            o_builder.Append(vo_e.Message);
            o_builder.Append('\n');
            o_builder.Append(vo_e.Data);
            o_builder.Append('\n');
            o_builder.Append(vo_e.StackTrace);
            Controller.displayErrorMessage(o_builder.ToString());
        }

		/// <summary>
		/// Utility method for displaying errors. Will display with an error header
		/// and an 'OK' button.
		/// </summary>
		/// <param name="vs_message">The message text to display.</param>
		internal static void displayErrorMessage(String vs_message)
		{
			displayMessage(vs_message, "Error!", MessageBoxButtons.OK);
		}

		/// <summary>
		/// Convience method for displaying messages.
		/// </summary>
		/// <param name="vs_message">The text of the message.</param>
		/// <param name="vs_title">Title of the message.</param>
		/// <param name="vo_buttons">The buttons to display.</param>
		internal static DialogResult displayMessage(String vs_message, String vs_title, MessageBoxButtons vo_buttons)
		{
			String s_msg = (vs_message != null && vs_message.Trim().Length > 0 ? vs_message : "<Missing Text>");
			String s_title = (vs_title != null && vs_title.Trim().Length > 0 ? vs_title : "");
            System.Console.WriteLine(vs_message);
			return System.Windows.Forms.MessageBox.Show(s_msg, "Binary Visualizer - " + s_title, vo_buttons);
		}

		///<summary>
		/// Open's a file chooser and allows the user to select a file.
		/// A null/empty file value can be returned (no validation is
		/// performed).
		/// <param name="vb_file_exists">boolean flag used to enable file existance checking</param>
		/// <returns>String</returns>
		/// </summary>
		internal static String getFileNameFromUser(bool vb_file_exists)
		{
			String s_file_name = null;

			FileDialog o_file_dialog = (vb_file_exists ? ((FileDialog)new OpenFileDialog()) : new SaveFileDialog());
			o_file_dialog.Title = "Select File";
			o_file_dialog.InitialDirectory = "c:\\";
			o_file_dialog.CheckFileExists = vb_file_exists;
			o_file_dialog.CheckPathExists = true;
			o_file_dialog.Filter = "All files(*.*)|*.*";
			o_file_dialog.FilterIndex = 1;
			o_file_dialog.RestoreDirectory = true;

			// if the user clicks ok, read the result, otherwise do nothing
			if (o_file_dialog.ShowDialog() == DialogResult.OK)
				s_file_name = o_file_dialog.FileName;

			return s_file_name;
		}

        /// <summary>
        /// Closes a Stream object and causes the reference to be null 
        /// even if there is an exception closing the stream. The pointer
        /// passed in as a argument will equal null after calling this method.
        /// </summary>
        /// <param name="vo_stream">Pointer/Reference to a Stream</param>
        private static void close(Stream vo_stream)
        {
            if (vo_stream != null)
            {
                try
                {
                    vo_stream.Close();
                }
                catch (Exception)
                {
                    //doesnt matter what the exception was
                }
                finally
                {
                    vo_stream = null;
                }
            }
        }
	}
}
