/*
 * Created by SharpDevelop.
 * User: Erik Dean
 * Date: 12/29/2007
 * Time: 8:46 PM
 * Copyright: 2007 USMA D/EECS ITOC. This code may be redistributed under the BSD License.
 * 
 * To change this template use Tools | Options | Coding | Edit Standard Headers.
 */

using System;
using System.IO;
using System.Collections.Generic;
using System.Drawing;
using System.Drawing.Imaging;
using System.Windows.Forms;

namespace VizBin
{
	/// <summary>
	/// Description of MainForm.
	/// </summary>
	public partial class MainForm : Form
	{
		// the controller we will be using
        Controller o_controller = new Controller();
		// the hex form to display
		HexForm o_hexform = null;
		// the search form to use
		SearchForm o_searchform;

		/// <summary>
		/// Creates the actual GUI form and starts the business logic executing for the program.
		/// </summary>
		public MainForm()
		{
			//
			// The InitializeComponent() call is required for Windows Forms designer support.
			//
			InitializeComponent();

            // fill the formats list
			VizBin.MainForm.fillFormatTable();
			object[] ao_name_objects = new object[ xas_names.Length ];
			for( int i = 0; i < xas_names.Length; i++ )
				ao_name_objects[i] = (object)xas_names[i];
			xo_format_select.Items.AddRange( ao_name_objects );

            // fill the image processor names
            String[] as_proc = Controller.getImageProcessorNames();
            object[] ao_proc_names = new object[as_proc.Length];
            System.Array.Copy(as_proc, ao_proc_names, as_proc.Length);
            xo_processors_select.Items.AddRange(ao_proc_names);
            xo_processors_select.SelectedIndex = (as_proc.Length > 0 ? 0 : -1);

            // set the default image output type to tiff
			xo_format_select.SelectedItem = "TIFF";

			// create the hexbox
			this.o_hexform = new HexForm(this);

			// create the searchform
			this.o_searchform = new SearchForm(this, o_controller);

			//diable the UI until a file is loaded
            this.disableUI();
		}
		
		/// <summary>
		/// Constructs the String to ImageFormat mapping.
		/// </summary>
		private static void fillFormatTable()
		{
			xo_names2formats.Clear();
			for(int i = 0; i < xas_names.Length; i++ )
				xo_names2formats.Add( xas_names[i], xao_formats[i] );
		}
		
		/// <summary>
		/// Utility method to disable the UI.
		/// </summary>
		private void disableUI()
		{
			setUIState(false);
		}
		
		/// <summary>
		/// Utility method to enable the UI.
		/// </summary>
		private void enableUI()
		{
			setUIState(true);
		}
		
		/// <summary>
		/// Enabled (true) or Disable (false) the user interface.
		/// </summary>
		/// <param name="vb_enabled">true to enabled, false to disable</param>
		private void setUIState(bool vb_enabled)
		{
			this.xo_format_select.Enabled = vb_enabled;
            this.xo_processors_select.Enabled = vb_enabled;
            this.io_btn_down.Enabled = this.o_controller.canMoveDown() && vb_enabled;
            this.io_btn_left.Enabled = this.o_controller.canMoveLeft() && vb_enabled;
            this.io_btn_up.Enabled = this.o_controller.canMoveUp() && vb_enabled;
            this.io_btn_right.Enabled = this.o_controller.canMoveRight() && vb_enabled;
		}

        /// <summary>
        /// Retrieve a String file path from the user and store it in the controller.
        /// </summary>
        /// <param name="sender">the calling object</param>
        /// <param name="e">the event that triggered this</param>
        private void loadFileToolStripMenuItem_Click(object sender, EventArgs e)
        {
            String s_file_name = Controller.getFileNameFromUser(true);
            if (o_controller.setInputFilename(s_file_name))
            {
				//update the titlebar so we know what we're looking at
				this.Text = "VizBin Win32 v0.1 - " + s_file_name;

				// tell the user we're searching
				this.io_info_lbl.Text = "Magic Search Results:\nSEARCH IN PROGRESS...";

                //update the image map
                this.updateMap();

                //update the actual image
                this.updateImage(DirectionalEnum.CENTER);

                //enable the UI now
                this.enableUI();
            }
        }

        /// <summary>
        /// update the image map
        /// </summary>
        private void updateMap()
        {
            Object o_selected = this.xo_processors_select.SelectedItem;
            if (o_selected != null && o_controller.isInputFileSet())
            {
                this.disableUI();
                AbstractImageProcessor o_proc = o_controller.getImageProcessor(o_selected.ToString().Trim());
                if (o_proc == null)
					throw new ArgumentException("Processor in UI list but has no value in AbstractImageProcessor map");

                // update the image map
                Image o_map = this.o_controller.getBinaryMap(this.imgMap.Width, this.imgMap.Height, o_proc);
                if (o_map != null)
                {
                    o_map = new Bitmap(o_map);
                    this.imgMap.Image = o_map;
                    this.imgMap.Refresh();
                }
                this.enableUI();
            }

        }

        /// <summary>
        /// Update the central image and move in the direction specified.
        /// </summary>
        /// <param name="ve_denum">teh directional enum specifying the direction to move in the file stream</param>
        private void updateImage(DirectionalEnum ve_denum)
        {
            Object o_selected = this.xo_processors_select.SelectedItem;

            if (o_selected != null && o_controller.isInputFileSet())
            {
                // prevent user interaction from overloading the app
                this.disableUI();

				// tell the user we're searching
				this.io_info_lbl.Text = "Magic Search Results:\nSEARCH IN PROGRESS...";

                AbstractImageProcessor o_proc = o_controller.getImageProcessor(o_selected.ToString().Trim());
                if (o_proc == null)
					throw new ArgumentException("Processor in UI list but has no value in AbstractImageProcessor map");

                //generate the bitmap map 
                Image o_img = this.o_controller.getImage(o_proc, ve_denum);
                if (o_img != null)
                {
                    o_img = new Bitmap(o_img);

                    // display the initial picture
                    this.imgBox.Image = o_img;
                    this.imgBox.Refresh();
                }

                // update the scaling factor label
                this.io_scaling_label.Text = this.o_controller.getScale();

				// update the hex form as well
				this.updateHexForm(ve_denum);

                // update the label chunk to display the current
                // position being viewed
                this.io_chunk_lbl.Text = this.o_controller.getOffsetDisplayString();

                // update the interesting things panel
                // ESD - BH
                //String s_results = this.o_controller.getMagicSearchResults();
                //this.io_info_lbl.Text = s_results;
                this.io_info_lbl.Text = "Magic Search Results: " + Environment.NewLine + "N/A";

                // re-enable the UI
                this.enableUI();
            }
        }

		/// <summary>
		/// Update the hex display for the given direction of movement in the image map.
		/// </summary>
		/// <param name="ve_denum">DirectionalEnum indicating which offsetstruct to read</param>
		private void updateHexForm(DirectionalEnum ve_denum)
		{
			if (o_controller.isInputFileSet() && this.o_hexform.Visible)
			{
				String s_hex = this.o_controller.getHex(ve_denum);
				this.o_hexform.setText(s_hex);
			}
		}

        /// <summary>
        /// Requests the application to exit.
        /// </summary>
        /// <param name="sender">ignored</param>
        /// <param name="e">ignored</param>
        private void exitToolStripMenuItem_Click(object sender, EventArgs e)
        {
            System.Windows.Forms.Application.Exit();
        }

        /// <summary>
        /// Move up in the image map
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void io_btn_up_Click(object sender, EventArgs e)
        {
            this.updateImage(DirectionalEnum.UP);
        }

        /// <summary>
        /// Move right in the image map
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void io_btn_right_Click(object sender, EventArgs e)
        {
            this.updateImage(DirectionalEnum.RIGHT);
        }

        /// <summary>
        /// Move down in the image map
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void io_btn_down_Click(object sender, EventArgs e)
        {
            this.updateImage(DirectionalEnum.DOWN);
        }

        /// <summary>
        /// Move left in the image map
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void io_btn_left_Click(object sender, EventArgs e)
        {
            this.updateImage(DirectionalEnum.LEFT);
        }

        /// <summary>
        /// Update the map and the image when the processor type changes
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void xo_processors_selectionChanged(object sender, EventArgs e)
        {
            this.updateMap();
            this.updateImage(DirectionalEnum.CENTER);
        }

		/// <summary>
		/// Hide or show the hex view depending on the selection
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void io_hex_toggle_Click(object sender, EventArgs e)
		{
			this.setHexFormVisible(this.io_hex_toggle.Checked);
		}

		/// <summary>
		/// Show or hide the hex view.
		/// </summary>
		/// <param name="vb_is_visible">true = show, false = hide</param>
		internal void setHexFormVisible(bool vb_is_visible)
		{
			o_hexform.Visible = vb_is_visible;
			io_hex_toggle.Checked = vb_is_visible;
			if (vb_is_visible)
				this.updateHexForm(DirectionalEnum.CENTER);
		}

		internal void setSearchFormVisible(bool vb_is_visible)
		{
			o_searchform.Visible = vb_is_visible;
		}

		/// <summary>
		/// Display the Search menu
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void findToolStripMenuItem_Click(object sender, EventArgs e)
		{
			this.setSearchFormVisible(true);
		}
	}


	internal class SearchForm : Form
	{
		// callback to the parent
		private MainForm io_mainform;
        private Controller io_controller;
		private TextBox io_txt_box = new TextBox();
		private TextBox io_txt_results = new TextBox();
		private Button io_srch_btn = new Button();
		private Label io_lbl = new Label();
		private TableLayoutPanel o_container = new TableLayoutPanel();

		/// <summary>
		/// Prevent people from calling the default constructor
		/// </summary>
		private SearchForm()
			: base()
		{
			// prevent outsiders from calling this constructor
		}

		/// <summary>
		/// Creates an instance of the searchform with a callback
		/// to the creator.
		/// </summary>
		/// <param name="vo_mainform">the parent gui</param>
		/// <param name="vo_controller">controller object that should be used for data manipulation</param>
		internal SearchForm(MainForm vo_mainform, Controller vo_controller)
			: this()
		{
			if (vo_mainform == null) throw new ArgumentException("MainForm cannot be null");
            if (vo_controller == null) throw new ArgumentException("Controller cannot be null");
            this.io_controller = vo_controller;
			this.io_mainform = vo_mainform;

			// do the form setup
			this.Text= "Find...";
			this.Name = "SearchForm";
			this.Icon = this.io_mainform.Icon;
			this.io_txt_box.Dock = DockStyle.Fill;

			// do the textbox setup
			FontFamily ff = new FontFamily("Times New Roman");
			Font font = new Font(ff, 12.0f, FontStyle.Regular);
			this.io_txt_box.Enabled = true;
			this.io_txt_box.ReadOnly = false;
			this.io_txt_box.ScrollBars = ScrollBars.None;
			this.io_txt_box.Multiline = false;
			this.io_txt_box.BorderStyle = BorderStyle.Fixed3D;
			this.io_txt_box.TextAlign = HorizontalAlignment.Left;
			this.io_txt_box.BackColor = Color.White;
			this.io_txt_box.Font = font;
			this.io_txt_box.Width = this.Width;

			// set up the results panel
			this.io_txt_results.Enabled = true;
			this.io_txt_results.ReadOnly = true;
			this.io_txt_results.ScrollBars = ScrollBars.Both;
			this.io_txt_results.Multiline = true;
			this.io_txt_results.Height = 100;
			this.io_txt_results.BorderStyle = BorderStyle.Fixed3D;
			this.io_txt_results.TextAlign = HorizontalAlignment.Left;
			this.io_txt_results.BackColor = Color.White;
			this.io_txt_results.Font = font;
			this.io_txt_results.Width = this.Width;

			// setup the layout manager
			this.AutoSize = true;
			this.o_container.AutoSize = true;
			this.o_container.ColumnCount = 2;
			this.o_container.RowCount = 10;

			// do the label setup
			this.io_lbl.Text = "Search For: ";
			this.io_lbl.AutoSize = true;

			// do the button setup
			this.io_srch_btn.Text = "Search";
			this.io_srch_btn.AutoSize = true;
			this.io_srch_btn.Click += new System.EventHandler(this.io_srch_btn_Click);

			// do the button setup
			this.Controls.Add(this.o_container);
			this.o_container.Controls.Add(this.io_lbl, 0, 0);
			this.o_container.Controls.Add(this.io_txt_box, 0, 1);
			this.o_container.Controls.Add(this.io_srch_btn, 0, 2);
			this.o_container.Controls.Add(this.io_txt_results, 0, 3);
			this.Size = this.o_container.Size;
		}

		/// <summary>
		/// Prevents the window from closing, just hides it.
		/// </summary>
		/// <param name="e">CancelEventArgs</param>
		protected override void OnClosing(System.ComponentModel.CancelEventArgs e)
		{
			e.Cancel = true;
			io_mainform.setSearchFormVisible(false);
		}

		protected void io_srch_btn_Click(Object sender, EventArgs e)
		{
			String s_txt = this.io_txt_box.Text;
            if (s_txt != null)
            {
                s_txt = s_txt.Trim();
                String s_result = this.io_controller.getSearchResults(s_txt);
				if (s_result != null && s_result.Length > 0)
					this.io_txt_results.Text = s_result;
				else
					this.io_txt_box.Text = "No Results Found.";
            }
		}
	}

	/// <summary>
	/// A simple form to display hex when desired by the user.
	/// </summary>
	internal class HexForm : Form
	{
		// the text box we're going to display the hex in
		private TextBox io_txt_box = new TextBox();
		private MainForm io_mainform;

		private HexForm() : base()
		{
			// this prevents outside classes from using this constructor
		}

		internal HexForm(MainForm vo_mainform) : this()
		{
			if (vo_mainform == null)
				throw new ArgumentException("Main Form cannot be null");

			// keep a reference to the main form
			this.io_mainform = vo_mainform;
			this.Size = new Size(800, 600);
			this.Text = "Byte Segment | Hex View | ASCII View";
			this.Name = "HexForm";
			this.Icon = this.io_mainform.Icon;

			// grid config
			this.io_txt_box.Dock = DockStyle.Fill;

			// do setup on the text area
			FontFamily ff = new FontFamily("Courier New");
			Font font = new Font(ff, 24.0f, FontStyle.Regular);
			this.io_txt_box.Enabled = true;
			this.io_txt_box.ReadOnly = true;
			this.io_txt_box.ScrollBars = ScrollBars.Both;
			this.io_txt_box.Multiline = true;
			this.io_txt_box.BorderStyle = BorderStyle.Fixed3D;
			this.io_txt_box.TextAlign = HorizontalAlignment.Left;
			this.Controls.Add(this.io_txt_box);
			this.io_txt_box.BackColor = Color.White;
			this.io_txt_box.Font = font;
		}

		/// <summary>
		/// Prevents the window from closing, just hides it.
		/// </summary>
		/// <param name="e">CancelEventArgs</param>
		protected override void  OnClosing(System.ComponentModel.CancelEventArgs e)
		{
			e.Cancel = true;
			io_mainform.setHexFormVisible(false);
		}

		/// <summary>
		/// Sets the text to display in our window.
		/// </summary>
		/// <param name="vs_text">string array of length 2, less will cause an exception, more will be ignored</param>
		internal void setText(String vs_text)
		{
			if (vs_text == null)
				throw new ArgumentException("String array cannot be null");
			this.io_txt_box.Text = vs_text;
		}
	}
}
