﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace VizBin
{
	/// <summary>
	/// A two dimensional table of row numbers and column nubmers representing an OffsetStruct element
	/// at each position.
	/// </summary>
	class OffsetTable
	{
		private IDictionary<Int32, IList<OffsetStruct>> IO_MAP = new Dictionary<Int32, IList<OffsetStruct>>();

		internal OffsetTable()
		{
			// not really used for anything but we need it to prevent this from being a public
			// constructor
		}

		internal void initTable(long vl_read_length, long vl_total_bytes)
		{
			if (vl_read_length < 1)
				throw new ArgumentException("Read length must be at least 1 byte");
			if (vl_total_bytes < 1)
				throw new ArgumentException("Total bytes must be at least 1 byte");

			// this is needed whenever this method is used to re-init an existing map
			IO_MAP.Clear();

			// find the total number of cells to put into the table
			long l_rem = vl_total_bytes % vl_read_length;
			long l_num_cells = vl_total_bytes / vl_read_length;
			if (l_rem > 0)
				l_num_cells++;

			// using the square root of the number of cells, we can get close to a grid
			// format in the table
			double d_sqrt = Math.Sqrt((double)l_num_cells);
			int n_rows_cols = (int)d_sqrt;

			// populate the table with the offset structs
			long l_curr_offset = 0;
			for (int rows = 0; rows < n_rows_cols; rows++)
			{
				for (int cols = 0; cols < n_rows_cols; cols++)
				{
					long l_next_offset = l_curr_offset + vl_read_length;
					OffsetStruct o_struct = new OffsetStruct(l_curr_offset, l_next_offset);
					l_curr_offset = l_next_offset;
					this.setCell(rows, cols, o_struct);
				}//endfor
			}//endfor

			// handle the extra offset values if there are some
			if (l_curr_offset < vl_total_bytes)
			{
				int n_next_row = n_rows_cols;
				long l_rem_bytes = vl_total_bytes - l_curr_offset;
				l_rem = l_rem_bytes % vl_read_length;
				long l_cols_to_create = (l_rem > 0 ? (l_rem_bytes / vl_read_length) + 1 : l_rem_bytes / vl_read_length);
				
				// create the remaining cols
				for (long l_col = 0; l_col < l_cols_to_create; l_col++)
				{
					// just a precaution to prevent adding more bytes to the cols than we should
					if (l_curr_offset >= vl_total_bytes)
						break;
					long l_next_offset = l_curr_offset + vl_read_length;
					if (l_next_offset > vl_total_bytes)
						l_next_offset = vl_total_bytes;
					this.setCell(n_next_row, (int)l_col, new OffsetStruct(l_curr_offset, l_next_offset));
					l_curr_offset = l_next_offset;
				}
			}
		}//endmethod

		/// <summary>
		/// Set the value of the specified cell (given the row and col) to the
		/// given struct value. ** ALL VALUES HERE ARE ZERO BASED **
		/// </summary>
		/// <param name="vn_row">The table row to update ** ZERO BASED **</param>
		/// <param name="vn_col">The table column to update ** ZERO BASED **</param>
		/// <param name="vo_struct">The OffsetStruct to insert</param>
		private void setCell(int vn_row, int vn_col, OffsetStruct vo_struct)
		{
			if (vn_row < 0)
				throw new ArgumentException("Row cannot be less 0");
			if (vn_col < 0)
				throw new ArgumentException("Col cannot be less than 0");
			if (vo_struct == null)
				throw new ArgumentException("OffsetStruct cannot be null");
			IList<OffsetStruct> o_struct_list = null;

			// if we have an existing row, grab that one, otherwise create a new one
			if (IO_MAP.ContainsKey(vn_row) && IO_MAP[vn_row] != null)
				o_struct_list = IO_MAP[vn_row];
			else
				o_struct_list = new List<OffsetStruct>();

			// set the item at the specified col
			o_struct_list.Insert(vn_col, vo_struct);

			// re-insert the row
			IO_MAP[vn_row] = o_struct_list;
		}//end method

		/// <summary>
		/// Returns the offset struct element at a specific "cell" in our dynamic
		/// table. 
	/// </summary>
		/// <param name="vn_row">the row number to retrieve</param>
		/// <param name="vn_col">the column number in that row</param>
		/// <returns>an offset struct or null</returns>
		internal OffsetStruct getCell( int vn_row, int vn_col )
		{
			if (vn_row < 0)
				throw new ArgumentException("Row cannot be less than 0");
			if (vn_col < 0)
				throw new ArgumentException("Col cannot be less than 0");
			if (vn_row > IO_MAP.Count)
				throw new ArgumentException("Row is greater than highest row number");
			
			OffsetStruct o_struct = null;
            if (IO_MAP.ContainsKey(vn_row))
            {
                IList<OffsetStruct> o_row = IO_MAP[vn_row];
                if (o_row != null && o_row.Count > vn_col)
                    o_struct = o_row[vn_col];
            }
			return o_struct;
		}//end method

		/// <summary>
		/// Returns the number of rows in the table.
		/// </summary>
		/// <returns>non-zero based number of rows in the current table</returns>
		internal int getRowCount()
		{
			return IO_MAP.Keys.Count;
		}

        /// <summary>
        /// Return the number of items in the given row. If the row doesn't exist,
        /// return 0. If the given row is less than 0, an ArgumentException is thrown.
        /// </summary>
        /// <param name="vn_row">the row to query</param>
        /// <returns>the number of columns in the row or 0 if the row doesnt exist</returns>
		internal int getColumnCount(int vn_row)
		{
            if (vn_row < 0)
                throw new ArgumentException("Row index cannot be less than 0");
            
            int n_row_count = 0;
            if (IO_MAP.ContainsKey(vn_row))
            {
                IList<OffsetStruct> o_list = IO_MAP[vn_row];
                if (o_list != null)
                    n_row_count = o_list.Count;
            }
            return n_row_count;
		}
	}//end class

	/// <summary>
	/// Defines an object with two long values, a start and an end value to be used for reading/writing
	/// parts of a binary value.
	/// </summary>
	class OffsetStruct
	{
		// the offset to start processing at
		private long il_offset_start = -1;
		// the offset to end processing at 
		private long il_offset_end = -1;

		/// <summary>
		/// Internal use only. This is kind of like an FOUO constructor.
		/// </summary>
		private OffsetStruct()
		{
			// this is private to prevent people from using it other than internally
		}

		/// <summary>
		/// Specifies the end offset, leaving the begining offset to be zero.
		/// </summary>
		/// <param name="vl_offset_end">set the end offset to this value</param>
		/// <exception cref="ArgumentException">If the end offset is less than 0</exception>
		internal OffsetStruct( long vl_offset_end ):this()
		{
			if( vl_offset_end < 0 )
				throw new ArgumentException("Offset end cannot be less than 0.");
			this.il_offset_end = vl_offset_end;
			this.il_offset_start = 0;
		}

		/// <summary>
		/// Creates an OffsetStruct object with the begining and offset specified.
		/// </summary>
		/// <param name="vl_offset_start">begining offset</param>
		/// <param name="vl_offset_end">end offset</param>
		/// <exception cref="ArgumentException">if the offset start is less than 0, if the end is less than 0, or if the start is less than the end</exception>
		internal OffsetStruct( long vl_offset_start, long vl_offset_end ):this(vl_offset_end)
		{
			if( vl_offset_start < 0 )
				throw new ArgumentException("Offset start cannot be less than 0.");
			if( vl_offset_end < vl_offset_start )
				throw new ArgumentException("Offset end ("+ vl_offset_end +") cannot be less than the start ("+ vl_offset_start +")");
			this.il_offset_start = vl_offset_start;
		}

		/// <summary>
		/// returns the start offset
		/// </summary>
		/// <returns>long >= 0</returns>
		internal long getStart()
		{
			return this.il_offset_start;
		}

		/// <summary>
		/// returns the end offset
		/// </summary>
		/// <returns>long &gt;= 0 &amp; long &gt;= start offset</returns>
		internal long getEnd()
		{
			return this.il_offset_end;
		}
	}
}
